# Copyright 2019 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import print_function
from __future__ import absolute_import

import time
import sys
import os
import ipaddr
import subprocess
import traceback
import stat
import shutil
import pwd
import grp
import tarfile
import ujson as json
import requests
from datetime import datetime
from netq_lib.common import utils, configuration
from netq_lib.common.exceptions import GetRequestError, PostRequestError
from netq_lib.common.enums import (
    ConfKw,
    DefaultPaths
)
from netq_lib.common.configuration import load_configuration, safe_load_all, is_opta


def error_handler(func):
    def wrapper(*args, **kwargs):
        self = args[0]
        string_io = sys.stdout  # sys.stdout is redirected to StringIO() in netqd
        for retry in range(0, OptaInstaller.RETRIES, 1):
            try:
                response = func(*args, **kwargs)
            except subprocess.CalledProcessError as ex:
                if retry < OptaInstaller.RETRIES - 1:
                    self.logger.warning('Retrying {} in {} seconds'.format(' '.join(ex.cmd), OptaInstaller.DELAY))
                    time.sleep(OptaInstaller.DELAY)
                else:
                    self.logger.error('{} failed to run'.format(' '.join(ex.cmd)))
                    self.logger.error(ex.output)
                    self.logger.error(string_io.getvalue())
                    self.unlock()
                    raise ex
            except (GetRequestError, PostRequestError) as ex:
                if retry < OptaInstaller.RETRIES - 1:
                    self.logger.warning('Retrying {} request to {} in {} seconds'.format(ex.method, ex.url,
                                                                                         OptaInstaller.DELAY))
                    time.sleep(OptaInstaller.DELAY)
                else:
                    self.logger.error(ex)
                    self.unlock()
                    raise ex
            except (ValueError, RuntimeError, KeyError, AssertionError) as ex:
                self.logger.error(ex)
                self.unlock()
                raise ex
            except Exception as ex:
                self.logger.error(ex)
                self.logger.error(traceback.format_exc())
                self.unlock()
                raise ex
            else:
                return response

    return wrapper


def check_server(func):
    def wrapper(*args, **kwargs):
        self = args[0]
        if self.is_pod_exists():
            return func(*args, **kwargs)
        else:
            raise ValueError('Kubernetes has to be initialized first')
    return wrapper


class OptaInstaller(object):
    ROOT_USER = 0
    KUBE_VERSION = ''  # determined at runtime
    KUBE_VERSION_MAJOR, KUBE_VERSION_MINOR = 1, 11
    PYTHON_MAJOR, PYTHON_MINOR = 2, 7
    DOCKER_VERSION = ''  # determined at runtime
    DOCKER_VERSION_MAJOR, DOCKER_VERSION_MINOR = 18, 6
    NETQ_MAJOR, NETQ_MINOR, NETQ_MICRO = 2, 2, 0
    OPTA_PATH = '/etc/app-release'
    RESOLV_PATH = '/etc/resolv.conf'
    AUTHORIZED_KEY_FILE_PATH = '/root/.ssh/authorized_keys'
    KUBE_FLANNEL_PATH = '/usr/share/cts-kubectl-config/kube-flannel.yml'
    REQUEST_HEADERS = {'accept': 'application/json', 'Content-Type': 'application/json'}
    DOCKER = '/usr/bin/docker'
    IP = '/sbin/ip'
    PING = '/bin/ping'
    GREP = '/bin/grep'
    SYSTEMCTL = '/bin/systemctl'
    KUBEADM = '/usr/bin/kubeadm'
    SSH_KEYGEN = '/usr/bin/ssh-keygen'
    DF = '/bin/df'
    NPROC = '/usr/bin/nproc'
    DOCKER_SERVICE_NAME = 'docker'
    KUBELET_SERVICE_NAME = 'kubelet'
    ADMIN_CONF = '/etc/kubernetes/admin.conf'
    NETQ_INSTALLER_APP_NAME = 'netq-installer'
    LOGS_APP_NAME = 'logs'
    DEFAULT_IFNAME = 'eth0'
    RETRIES = 12
    DELAY = 10
    HEADER = 'opta-installer'

    def __init__(self, connection):
        self.connection = connection
        self.config = configuration.get_config(ConfKw.NETQD)
        self.logger = utils.get_logger(ConfKw.NETQD, self.config)
        self.ifname_addresses = dict()
        self.data_path = '/mnt'
        self.pod_network_cidr = '10.244.0.0/16'
        self.node_name = 'cumulus.netq'
        self.scratch_path = '/tmp'
        self.kubectl = '/usr/bin/kubectl'
        self.installables_dir = '{}/installables'.format(self.data_path)
        self.netq_installer_scratch_dir = '{}/installer'.format(self.scratch_path)
        self.docker_images_dir = '{}/images'.format(self.netq_installer_scratch_dir)
        self.fluentd_spec_dir = '{}/logs'.format(self.netq_installer_scratch_dir)
        self.installer_spec_dir = '{}/netq-installer'.format(self.netq_installer_scratch_dir)
        self.tarball_path = ''

    def print_out(self, msg):
        msg = '{}: {}: {}'.format(datetime.utcnow(), OptaInstaller.HEADER, msg)
        utils.print_out(self.connection, msg)
        
    def log_msg(self, msg):
        msg = '{}: {}'.format(OptaInstaller.HEADER, msg)
        self.logger.info(msg)

    def lock(self):
        if os.path.exists(DefaultPaths.INSTALLER_LOCK):
            raise OSError('{} lock in use'.format(DefaultPaths.INSTALLER_LOCK))
        else:
            msg = 'OPTA installer executing...Blocking all other CLI commands'
            self.logger.warning(msg)
            open(DefaultPaths.INSTALLER_LOCK, 'w').close()

    def unlock(self):
        if os.path.exists(DefaultPaths.INSTALLER_LOCK):
            msg = 'OPTA installer finished...Unblocking all other CLI commands'
            self.logger.warning(msg)
            os.remove(DefaultPaths.INSTALLER_LOCK)

    def set_kubectl(self, server='127.0.0.1', config='/etc/kubernetes/admin.conf'):
        if not os.path.exists(config):
            raise ValueError('Kubernetes has to be initialized first')

        if server == '127.0.0.1':
            rc, kube_config = load_configuration(OptaInstaller.ADMIN_CONF)
            if kube_config is None:
                kube_config = dict()
            url = kube_config.get('clusters', [dict()])[0].get('cluster', dict()).get('server')
            if url is None:
                raise ValueError('Kubernetes has to be initialized first')
            server = url[8:20]

        os.chmod(config, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
        self.kubectl = '/usr/bin/kubectl --server=https://{}:6443 --kubeconfig={}'.format(server, config)
        return self.kubectl

    def get_service_cluster_ip(self, service_name):
        msg = 'Getting cluster IP for service {}'.format(service_name)
        self.log_msg(msg)
        cmd = self.kubectl.split() + ['get', 'service', service_name, '-o', 'jsonpath=\"{.spec.clusterIP}\"']
        return subprocess.check_output(cmd, stderr=subprocess.STDOUT).strip('\"')

    @error_handler
    def package_versions(self):
        msg = 'Checking for Python {}.{}'.format(OptaInstaller.PYTHON_MAJOR, OptaInstaller.PYTHON_MINOR)
        self.log_msg(msg)
        assert sys.version_info.major == OptaInstaller.PYTHON_MAJOR and \
               sys.version_info.minor == OptaInstaller.PYTHON_MINOR, \
            'Python {}.{} required'.format(OptaInstaller.PYTHON_MAJOR, OptaInstaller.PYTHON_MINOR)

        msg = 'Checking for Kubernetes >= {}.{}'.format(OptaInstaller.KUBE_VERSION_MAJOR,
                                                        OptaInstaller.KUBE_VERSION_MINOR)
        self.log_msg(msg)
        output = subprocess.check_output('{} version --client --output=json'.format(self.kubectl).split(),
                                         stderr=subprocess.STDOUT)
        kubectl_info = json.loads(output)
        version_json = kubectl_info.get('clientVersion')
        assert int(version_json.get('major')) >= OptaInstaller.KUBE_VERSION_MAJOR and \
               int(version_json.get('minor')) >= OptaInstaller.KUBE_VERSION_MINOR, \
            'Kubernetes >= {}.{} required'.format(OptaInstaller.KUBE_VERSION_MAJOR, OptaInstaller.KUBE_VERSION_MINOR)
        OptaInstaller.KUBE_VERSION = version_json.get('gitVersion')

        msg = 'Checking for Docker >= {}.{}'.format(OptaInstaller.DOCKER_VERSION_MAJOR,
                                                    OptaInstaller.DOCKER_VERSION_MINOR)
        self.log_msg(msg)
        output = subprocess.check_output('{} --version'.format(OptaInstaller.DOCKER).split(),
                                         stderr=subprocess.STDOUT)

        # sample output: Docker version 18.06.1-ce, build e68fc7a
        version_info = output.split()[2]
        major, minor = int(version_info.split('.')[0]), int(version_info.split('.')[1])
        assert major >= OptaInstaller.DOCKER_VERSION_MAJOR and minor >= OptaInstaller.DOCKER_VERSION_MINOR, \
            'Docker >= {}.{} required'.format(OptaInstaller.DOCKER_VERSION_MAJOR, OptaInstaller.DOCKER_VERSION_MINOR)
        OptaInstaller.DOCKER_VERSION = version_info

        netq_version = '{}.{}.{}'.format(OptaInstaller.NETQ_MAJOR, OptaInstaller.NETQ_MINOR, OptaInstaller.NETQ_MICRO)
        msg = 'Checking for NetQ {}'.format(netq_version)
        self.log_msg(msg)
        if not os.path.exists(OptaInstaller.OPTA_PATH):
            raise ValueError('{} does not exist'.format(OptaInstaller.OPTA_PATH))
        with open(OptaInstaller.OPTA_PATH, 'r') as fd:
            for line in fd:
                line_list = line.strip().split('=')
                if len(line_list) > 1 and line_list[0] == 'APPLIANCE_VERSION':
                    major, minor, micro = tuple(line_list[1].split('.'))
                    assert int(major) >= OptaInstaller.NETQ_MAJOR and \
                           int(minor) >= OptaInstaller.NETQ_MINOR and \
                           int(micro) >= OptaInstaller.NETQ_MICRO, 'NetQ {} required'.format(netq_version)
                    break

    @error_handler
    def create_dirs(self):
        msg = 'Creating directories'
        self.log_msg(msg)
        for dir in ['zookeeper/data', 'zookeeper/log', 'cassandra', 'kafka', 'prometheus', 'var/lib/docker',
                    'netq-app-search/index', 'box']:
            path = '{}/{}'.format(self.data_path, dir)
            if not os.path.exists(path):
                msg = 'Creating {}'.format(path)
                self.log_msg(msg)
                os.makedirs(path)
                os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)

        netq_log_path = '/var/log/netq'
        if not os.path.exists(netq_log_path):
            msg = 'Creating {}'.format(netq_log_path)
            self.log_msg(msg)
            os.makedirs(netq_log_path)

        # Creating dir for authorized key installation after ssh key generation
        ssh_root_path = '/root/.ssh'
        if not os.path.exists(ssh_root_path):
            msg = 'Creating {}'.format(ssh_root_path)
            self.log_msg(msg)
            os.makedirs(ssh_root_path)

    @error_handler
    def create_installer_scratch_dirs(self):
        if not os.path.exists(self.netq_installer_scratch_dir):
            msg = 'Creating {}'.format(self.netq_installer_scratch_dir)
            self.log_msg(msg)
            os.makedirs(self.netq_installer_scratch_dir)
            os.chmod(self.netq_installer_scratch_dir, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)

    @error_handler
    def extract_tarball(self, tarball_path):
        if os.path.exists(self.netq_installer_scratch_dir):
            msg = "Cleaning up installer scratch dir"
            self.log_msg(msg)

            shutil.rmtree(self.netq_installer_scratch_dir)
        self.create_installer_scratch_dirs()
        msg = "Extracting tar ball {} at {}".format(tarball_path, self.netq_installer_scratch_dir)
        self.log_msg(msg)
        tar = tarfile.open(tarball_path)
        tar.extractall(self.netq_installer_scratch_dir)
        tar_balls = [x for x in os.listdir(self.netq_installer_scratch_dir) if x.endswith(".tgz")]
        for tar_ball in tar_balls:
            tar = tarfile.open("{}/{}".format(self.netq_installer_scratch_dir, tar_ball))
            tar.extractall(self.netq_installer_scratch_dir)
        tar.close()

    @error_handler
    def interface_address(self, ifname):
        msg = 'Checking for {} IP address assignment'.format(ifname)
        self.log_msg(msg)
        try:
            out = subprocess.check_output('{} -4 -br addr show {}'.format(OptaInstaller.IP, ifname).split(),
                                          stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError:
            raise ValueError('undefined interface {}'.format(ifname))
        else:
            if len(out):
                ip_address = ipaddr.IPNetwork(out.split()[2])
                try:
                    msg = 'Checking if {} is reachable'.format(ip_address.ip)
                    self.log_msg(msg)
                    subprocess.check_output('{} {} -c 3'.format(OptaInstaller.PING, ip_address.ip).split(),
                                            stderr=subprocess.STDOUT)
                except subprocess.CalledProcessError:
                    raise ValueError('IP address {} is not reachable'.format(ip_address.ip))
                else:
                    return ip_address.ip
            else:
                msg = 'IP address not found for interface: {}'.format(ifname)
                self.log_msg(msg)
                raise ValueError('Failed to find assigned IP address for {}'.format(ifname))

    @error_handler
    def check_service(self, service_name):
        msg = 'Checking for {} service status'.format(service_name)
        self.log_msg(msg)
        try:
            out = subprocess.check_output(
                '{} show -p SubState -p ActiveState -p LoadState {}'.format(OptaInstaller.SYSTEMCTL,
                                                                            service_name).split(),
                stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError:
            raise ValueError('{} service is not running'.format(service_name))
        else:
            out_list = out.strip().split("\n")
            out_dict = dict(s.split('=') for s in out_list)
            if out_dict['LoadState'] != 'loaded':
                msg = "{} Service is not loaded".format(service_name)
                self.log_msg(msg)
                return False
            elif out_dict['ActiveState'] != 'active':
                msg = '{} service is not in active state'.format(service_name)
                self.log_msg(msg)
                return False
            elif out_dict['SubState'] != 'running':
                msg = '{} service is not running'.format(service_name)
                self.log_msg(msg)
                return False
            else:
                msg = '{} service is running'.format(service_name)
                self.log_msg(msg)
                return True

    @error_handler
    def start_service(self, service_name):
        msg = "Starting {} service".format(service_name)
        self.log_msg(msg)
        cmd = '{} start {}'.format(OptaInstaller.SYSTEMCTL, service_name)
        try:
            subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError:
            raise ValueError('Error occurred in starting service {}'.format(service_name))
        else:
            if self.check_service(service_name):
                msg = '{} Service Started'.format(service_name)
                self.log_msg(msg)
            else:
                raise RuntimeError('Error occurred while starting {} service '.format(service_name))

    @error_handler
    def init_kube_cluster(self, ifname_address):
        msg = 'Initializing kubernetes cluster'
        self.log_msg(msg)
        subprocess.check_output('{} --pod-network-cidr={} --apiserver-advertise-address={} '
                                '--kubernetes-version={} --node-name={} init'.format(OptaInstaller.KUBEADM,
                                                                                     self.pod_network_cidr,
                                                                                     ifname_address,
                                                                                     OptaInstaller.KUBE_VERSION,
                                                                                     self.node_name).split(),
                                stderr=subprocess.STDOUT)

    @error_handler
    def check_resolv_file(self):
        msg = 'Creating resolv config file'
        self.log_msg(msg)
        if not os.path.exists(OptaInstaller.RESOLV_PATH):
            open(OptaInstaller.RESOLV_PATH, 'a').close()

    @error_handler
    def check_authorized_keys_file(self):
        msg = 'Checking and creating authorized_key file for ssh authorization'
        self.log_msg(msg)
        if not os.path.exists(OptaInstaller.AUTHORIZED_KEY_FILE_PATH):
            open(OptaInstaller.AUTHORIZED_KEY_FILE_PATH, 'a').close()

    @error_handler
    def setup_kube_admins(self, ifname):
        for username, home_dir in [('cumulus', '/home/cumulus'), ('root', '/root')]:
            msg = 'Assigning {} as Kubernetes admins'.format(username)
            self.log_msg(msg)
            if os.path.exists(home_dir):
                path = '{}/.kube'.format(home_dir)
                if os.path.exists(path):
                    shutil.rmtree(path)
                if not os.path.exists(path):
                    os.makedirs(path)
                if not os.path.exists('/etc/kubernetes/admin.conf'):
                    raise ValueError('/etc/kubernetes/admin.conf does not exist')

                self.set_kubectl(server=self.ifname_addresses.get(ifname))
                shutil.copy('/etc/kubernetes/admin.conf', '{}/.kube/config'.format(home_dir))
                os.chown('{}/.kube/config'.format(home_dir),
                         pwd.getpwnam(username).pw_uid, grp.getgrnam(username).gr_gid)

    @error_handler
    def check_kube_flannel(self):
        msg = 'Verifying kubernetes flannel'
        self.log_msg(msg)
        rc, yaml_config = safe_load_all(OptaInstaller.KUBE_FLANNEL_PATH)
        net_conf = yaml_config[3].get('data').get('net-conf.json')
        ip = json.loads(net_conf).get('Network')
        assert ip == self.pod_network_cidr, \
            'Pod Network CIDR {} does not match with {} in {}'.format(self.pod_network_cidr, ip,
                                                                      OptaInstaller.KUBE_FLANNEL_PATH)

    @error_handler
    def install_kube_flannel(self):
        msg = 'Installing kubernetes flannel'
        self.log_msg(msg)
        subprocess.check_output('sudo -u cumulus {} apply -f {}'.format(self.kubectl,
                                                                        OptaInstaller.KUBE_FLANNEL_PATH).split(),
                                stderr=subprocess.STDOUT)

    @error_handler
    def install_kube_cluster(self):
        msg = 'Installing kubernetes cluster'
        self.log_msg(msg)
        subprocess.check_output('sudo -u cumulus {} taint node cumulus.netq '
                                'node-role.kubernetes.io/master:NoSchedule-'.format(self.kubectl).split(),
                                stderr=subprocess.STDOUT)

    @error_handler
    def check_kube_cluster(self):
        msg = 'Validating kubernetes cluster'
        self.log_msg(msg)
        cmd = '{} cluster-info'.format(self.kubectl)
        subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT)

    @error_handler
    def load_docker_images(self):
        msg = 'Loading docker images from tarball extract'
        self.log_msg(msg)
        images = [x for x in os.listdir(self.docker_images_dir) if x.endswith(".tar")]
        for image in images:
            cmd = '{} load -i {}/images/{}'.format(OptaInstaller.DOCKER, self.netq_installer_scratch_dir, image)
            subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT)

    @error_handler
    def default_route(self):
        msg = 'Checking for default route'
        self.log_msg(msg)
        out = subprocess.check_output('{} route show default 0.0.0.0/0'.format(OptaInstaller.IP).split(),
                                      stderr=subprocess.STDOUT)
        if len(out) == 0:
            raise ValueError('no default route found')

    @error_handler
    def deploy_k8_pods(self, pod_spec_dir, app_name):
        msg = 'Setting up {}'.format(app_name)
        self.log_msg(msg)
        yml_files = [x for x in os.listdir(pod_spec_dir) if x.endswith(".yml")]
        for yml in yml_files:
            cmd = '{} create -f {}/{}'.format(self.kubectl, pod_spec_dir, yml)
            subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT)
            if yml == 'persistent-vol.yml':
                time.sleep(20)
            elif yml == 'pv-claim.yml':
                time.sleep(5)

    @error_handler
    def ssh_keygen(self):
        msg = 'Generating SSH keys'
        self.log_msg(msg)

        key_name = 'installer-ssh-keys'
        cmd = '{} get secrets -o json'.format(self.kubectl)
        output = json.loads(subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT))
        for secret in output.get('items', []):
            if secret.get('metadata', dict()).get('name') == key_name:
                cmd = '{} delete secret {}'.format(self.kubectl, key_name)
                subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT)
                break

        for ssh_key in ['id_rsa', 'id_rsa.pub']:
            path = '{}/{}'.format(self.scratch_path, ssh_key)
            if os.path.exists(path):
                os.remove(path)
        subprocess.check_output([OptaInstaller.SSH_KEYGEN, '-t', 'rsa',
                                 '-f', '{}/id_rsa'.format(self.scratch_path), '-N', ''],
                                stderr=subprocess.STDOUT)
        with open('{}/id_rsa.pub'.format(self.scratch_path), 'r') as id_rsa, \
                open('/root/.ssh/authorized_keys', 'a') as auth_keys:
            for line in id_rsa:
                auth_keys.write(line)
        os.chmod('/root/.ssh/authorized_keys', stat.S_IRUSR | stat.S_IWUSR)
        msg = 'Generating kubernetes secret'
        self.log_msg(msg)
        cmd = '{} create secret generic {} --from-file={}/id_rsa'.format(self.kubectl, key_name, self.scratch_path)
        subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT)

    @error_handler
    def get_pod_name(self, app_name):
        msg = 'Getting pod name for app name {}'.format(app_name)
        self.log_msg(msg)
        cmd = '{} get pod -l app={} -o jsonpath="{}"'.format(self.kubectl, app_name, "{.items[0].metadata.name}")
        out = subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT).strip('\"')
        if out.startswith(app_name):
            msg = 'Pod Name from app name {}: {} '.format(app_name, out)
            self.log_msg(msg)
            return str(out)
        else:
            self.logger.error("Error occurred while getting pod name from app name {}".format(app_name))

    @error_handler
    def validate_installer(self, timeout=60):
        msg = 'Validating installer. Polling again in {} seconds.'.format(OptaInstaller.DELAY)
        self.log_msg(msg)
        cmd = self.kubectl.split() + ['get', 'pod', '-l', 'app=netq-installer', '-o',
                                      'jsonpath=\"{.items[0].status.containerStatuses[0].state}\"']
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        installer_url = 'http://127.0.0.1:32500/netq/installer/v1/job'
        try:
            response = requests.get(url=installer_url, headers=OptaInstaller.REQUEST_HEADERS, timeout=timeout)
        except requests.exceptions.RequestException:
            raise GetRequestError(url=installer_url, headers=str(OptaInstaller.REQUEST_HEADERS))
        else:
            if response.status_code != requests.codes['ok']:
                raise GetRequestError(url=installer_url, headers=str(OptaInstaller.REQUEST_HEADERS),
                                      response=str(response.json()))

    @error_handler
    def validate_installation(self, timeout=60):
        opta_ip = self.get_service_cluster_ip('netq-opta')
        opta_health_check_url = 'http://{}:8080/health'.format(opta_ip)
        try:
            response = requests.get(url=opta_health_check_url, headers=OptaInstaller.REQUEST_HEADERS, timeout=timeout)
        except requests.exceptions.RequestException:
            raise GetRequestError(url=opta_health_check_url, headers=str(OptaInstaller.REQUEST_HEADERS))
        else:
            response_json = response.json()
            if response_json.get('isHealthy'):
                msg = 'OPTA is healthy and running'
                self.log_msg(msg)
            else:
                self.logger.error(response_json)
                raise ValueError('OPTA is not healthy')

    @error_handler
    def installer_api(self, tarball_path, upgrade=False, timeout=60):
        msg = 'Calling installer API'
        self.log_msg(msg)
        job_installer_url = 'http://127.0.0.1:32500/netq/installer/v1/job'
        data = json.dumps({'is_installer_upgrade': upgrade, 'path': tarball_path})
        try:
            response = requests.post(url=job_installer_url, headers=OptaInstaller.REQUEST_HEADERS, timeout=timeout,
                                     data=data)
        except requests.exceptions.RequestException:
            raise PostRequestError(url=job_installer_url, headers=str(OptaInstaller.REQUEST_HEADERS), data=data)
        else:
            if response.status_code != requests.codes['ok']:
                raise PostRequestError(url=job_installer_url, headers=str(OptaInstaller.REQUEST_HEADERS),
                                       data=data, response=str(response.json()))

    @error_handler
    def poll_kube_jobs(self, minutes, timeout=60):
        installer_url = 'http://127.0.0.1:32500/netq/installer/v1/job'
        secs = 60

        for poll in range(0, minutes, 1):
            try:
                response = requests.get(url=installer_url,
                                        headers=OptaInstaller.REQUEST_HEADERS, timeout=timeout)
            except requests.exceptions.RequestException:
                ex = GetRequestError(url=installer_url, headers=str(OptaInstaller.REQUEST_HEADERS))
                self.logger.error(ex)
                raise ex
            else:
                if response.status_code != requests.codes['ok']:
                    ex = GetRequestError(url=installer_url, headers=str(OptaInstaller.REQUEST_HEADERS))
                    self.logger.error(ex)
                    raise ex

                kube_jobs = response.json().get('jobs', dict())
                if len(kube_jobs):
                    msg = 'Waiting for Kubernetes jobs to complete. Polling again in {} seconds'.format(secs)
                    self.log_msg(msg)
                    time.sleep(secs)
                else:
                    msg = 'Checking kubernetes installer job status'
                    self.log_msg(msg)
                    self.check_kube_job_status()
                    return

    def check_kube_job_status(self):
        pod_name = self.get_pod_name(OptaInstaller.NETQ_INSTALLER_APP_NAME)

        try:
            # check the file cmdstat.id exists or not.
            cmd = self.kubectl.split() + ['exec', pod_name, '--', 'bash', '-c', 'test -f /cmdstat.id']
            subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError:
            # logic copied over from netq-k8-deploy/netq-install.sh
            return
        else:
            try:
                # check the content of cmdstat.d file
                cmd = self.kubectl.split() + ['exec', pod_name, '--', 'bash', '-c', 'cat /cmdstat.id']
                subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError:
                msg = 'Kubernetes job failed: \'kubectl logs {}\' for more details.'.format(pod_name)
                raise RuntimeError(msg)
            else:
                return

    @error_handler
    @check_server
    def set_kube_env_var(self, key, value):
        msg = 'Setting Kubernetes environment variable {}={}'.format(key, value)
        self.log_msg(msg)
        subprocess.check_output('{} set env deployment.extensions/netq-opta-deploy {}={}'.format(
            self.kubectl, key, value).split(), stderr=subprocess.STDOUT)

    @error_handler
    def check_installer_dir(self):
        msg = 'Checking for installer directory'
        self.log_msg(msg)
        path = '{}/installables'.format(self.data_path)
        if not os.path.exists(path):
            raise ValueError('{} installer directory does not exist'.format(path))

    @error_handler
    def check_ansible(self):
        msg = 'Checking ansible connectivity'
        self.log_msg(msg)
        pod_name = ''
        out = subprocess.check_output('{} get pods -o json'.format(self.kubectl).split(), stderr=subprocess.STDOUT)
        response_json = json.loads(out)
        for pod in response_json.get('items'):
            metadata = pod.get('metadata')
            if metadata.get('name').startswith('netq-installer'):
                pod_name = str(metadata.get('name'))
                break

        cmd = self.kubectl.split() + ['exec', pod_name, '--', 'bash', '-c', 'ansible all -m ping']
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)

    @error_handler
    def copy_netq_config(self):
        msg = 'Copying netq config file'
        self.log_msg(msg)
        shutil.copy('/usr/share/cts-kubectl-config/netq.yml', '/etc/netq/config.d/netq.yml')

    @error_handler
    def restart_services(self):
        subprocess.check_output('{} reset-failed'.format(OptaInstaller.SYSTEMCTL).split(),
                                stderr=subprocess.STDOUT)
        for service in ['rsyslog.service']:
            msg = 'Restarting {}'.format(service)
            self.log_msg(msg)
            subprocess.check_output('{} restart {}'.format(OptaInstaller.SYSTEMCTL, service).split(),
                                    stderr=subprocess.STDOUT)
            time.sleep(5)
            subprocess.check_output('{} status {}'.format(OptaInstaller.SYSTEMCTL, service).split(),
                                    stderr=subprocess.STDOUT)

    @error_handler
    def ram_requirements(self):
        msg = 'Checking minimum RAM requirements'
        self.log_msg(msg)
        path = '/proc/meminfo'
        if not os.path.isfile(path):
            raise ValueError('{} does not exist'.format(path))
        out = subprocess.check_output('{} MemTotal: {}'.format(OptaInstaller.GREP, path).split(),
                                      stderr=subprocess.STDOUT)
        mem_total = out.split()[1]

        min_mem_8gb = 8000000
        if int(mem_total) < min_mem_8gb:
            raise ValueError('minimum of 8 GB RAM required')

    @error_handler
    def cpu_requirements(self):
        msg = 'Checking for minimum CPU requirements'
        self.log_msg(msg)
        out = subprocess.check_output([OptaInstaller.NPROC], stderr=subprocess.STDOUT)
        min_cores = 4
        num_cores = int(out)
        if num_cores < min_cores:
            raise ValueError('minimum of {} CPU cores required'.format(min_cores))

    @error_handler
    def disk_requirements(self):
        msg = 'Checking for minimum disk requirements'
        self.log_msg(msg)
        min_disk_256gb_in_kb = 256000000
        out = subprocess.check_output('{} --output=size,target'.format(OptaInstaller.DF).split(),
                                      stderr=subprocess.STDOUT)
        for line in out.split('\n'):
            line_list = line.strip().split()
            if len(line_list) > 1 and line_list[1] == '/':
                if int(line_list[0]) < min_disk_256gb_in_kb:
                    raise ValueError('minimum of {} GB disk space required'.format(int(line_list[0])))
                else:
                    break

    @error_handler
    def purge_opta(self):
        msg = 'Resetting OPTA'
        self.log_msg(msg)
        subprocess.check_output('{} reset -f'.format(OptaInstaller.KUBEADM).split(), stderr=subprocess.STDOUT)
        path = '/home/cumulus/.kube/config'
        if os.path.exists(path):
            os.remove(path)

        path = '/var/lib/cni/networks/cbr0'
        if os.path.exists(path):
            for root, dirs, files in os.walk(path):
                for f in files:
                    os.remove(os.path.join(root, f))
                for d in dirs:
                    shutil.rmtree(os.path.join(root, d))

    @error_handler
    def overwrite_defaults(self, file_path):
        msg = 'Overwriting default values with {}'.format(file_path)
        self.log_msg(msg)
        rc, current_config = load_configuration(file_path)
        for key in ['data-dir', 'pod-network-dir', 'node-name', 'scratch-dir']:
            if key not in current_config.keys():
                raise KeyError('Missing entry for {} in {}'.format(key, file_path))

        self.data_path = current_config.get('data-dir')
        self.pod_network_cidr = current_config.get('pod-network-dir')
        self.node_name = current_config.get('node-name')
        self.scratch_path = current_config.get('scratch-dir')
        # todo check for only_opta config variable as well

    @error_handler
    def check_tarball_path(self, tarball):
        if len(tarball.split('/')) == 1:
            path = '{}/{}'.format(self.installables_dir, tarball)
            if not os.path.exists(path):
                raise ValueError('{} does not exist'.format(path))
        else:
            subdir = tarball.split('/')
            if len(subdir) == 4 and subdir[1] == 'mnt' and subdir[2] == 'installables':
                path = tarball
            else:
                raise ValueError('Tarball must be under {}'.format(self.installables_dir))

        self.log_msg('Checking tarball path {}'.format(path))
        if 'opta' in tarball.lower() or 'cloud' in tarball.lower():
            return path
        else:
            raise ValueError('{} is an invalid appliance name'.format(tarball))

    def is_pod_exists(self):
        if not os.path.exists(OptaInstaller.ADMIN_CONF):
            return False
        if len(self.kubectl.split()) == 1:
            self.set_kubectl()
        try:
            subprocess.check_output('{} get pods'.format(self.kubectl).split())
        except subprocess.CalledProcessError:
            return False
        else:
            return True

    @check_server
    def upgrade(self, tarball):
        if not is_opta():
            raise ValueError('Must be on OPTA to run command')

        self.lock()
        self.print_out('Upgrading OPTA')
        self.tarball_path = self.check_tarball_path(tarball)
        self.ssh_keygen()

        self.log_msg('Upgrading netq-installer pod')
        self.installer_api(tarball_path=self.tarball_path, upgrade=True)
        self.poll_kube_jobs(minutes=30)

        self.log_msg('Upgrading netq-opta pod')
        self.installer_api(tarball_path=self.tarball_path, upgrade=False)
        self.poll_kube_jobs(minutes=30)

        self.print_out('Validating installation')
        self.validate_installation()
        self.restart_services()
        self.unlock()

    def install(self, key, ifname, tarball, file_path='', force=False):
        if not is_opta():
            msg = 'Must be on OPTA to run command'
            self.logger.error(msg)
            raise ValueError(msg)

        if not force and self.is_pod_exists():
            msg = 'Kubernetes pod found. Run OPTA installation command with argument \'force\' to reset OPTA'
            self.logger.warning(msg)
            raise ValueError(msg)

        self.lock()
        if len(file_path):
            self.overwrite_defaults(file_path)

        self.check_installer_dir()
        self.print_out('Checking hardware requirements')
        self.ram_requirements()
        self.cpu_requirements()
        # self.disk_requirements()

        self.print_out('Checking package requirements')
        self.package_versions()
        self.ifname_addresses[ifname] = self.interface_address(ifname)
        self.default_route()
        self.create_dirs()
        self.tarball_path = self.check_tarball_path(tarball)

        self.print_out('Resetting OPTA')
        self.purge_opta()

        self.print_out('Extracting tarball {}'.format(self.tarball_path))
        self.extract_tarball(self.tarball_path)
        self.check_resolv_file()
        self.check_authorized_keys_file()
        is_docker_service_running = self.check_service(self.DOCKER_SERVICE_NAME)
        if not is_docker_service_running:
            self.start_service(self.DOCKER_SERVICE_NAME)

        self.print_out('Initializing kubernetes cluster')
        self.init_kube_cluster(self.ifname_addresses.get(ifname))
        self.setup_kube_admins(ifname)
        is_kubelet_service_running = self.check_service(self.KUBELET_SERVICE_NAME)
        if not is_kubelet_service_running:
            self.start_service(self.KUBELET_SERVICE_NAME)
        self.check_kube_flannel()
        self.install_kube_flannel()
        self.install_kube_cluster()
        self.check_kube_cluster()
        self.load_docker_images()

        self.print_out('Deploying kubernetes pods')
        self.deploy_k8_pods(self.fluentd_spec_dir, self.LOGS_APP_NAME)
        self.ssh_keygen()
        self.deploy_k8_pods(self.installer_spec_dir, self.NETQ_INSTALLER_APP_NAME)

        self.print_out('Validating installation')
        self.validate_installer()
        self.check_ansible()
        self.installer_api(self.tarball_path)
        self.poll_kube_jobs(minutes=30)
        self.set_kube_env_var('OPTA_CONFIG_KEY', key)
        self.validate_installation()
        # todo skipping this step for a while
        # self.copy_netq_config()
        self.restart_services()
        self.unlock()

